package ffmpegconverter;

import ffmpegconverter.converter.AnimeConverter;
import ffmpegconverter.converter.AudioConverter;
import ffmpegconverter.converter.Converter;
import ffmpegconverter.converter.VideoConverter;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.function.Predicate;

/**
 * Converts a files or files in a folder to a web playable mp4.
 */
public class Main {
    private static final String FFPROBE_PATH = "ffprobe"; //Can be just ffprobe if it's in the path
    private static final String FFMPEG_PATH = "ffmpeg"; //Can be just ffmpeg if it's in the path
    private static final Scanner READER = new Scanner(System.in, "UTF-8");
    private static final BufferedWriter WRITER = new BufferedWriter(new OutputStreamWriter(System.out));
    private static Converter con = null;

    public static void main(String[] args) throws IOException {
        //System.out.println(tokenizer("AnimeConverter -audiolang jap,eng -sublang eng,nor,* \"C:\\Users\\Kristian\\Downloads\\Anime\\[Kametsu] ERASED (BD 1080p Hi10 FLAC)\""));
        //parser(tokenizer("AnimeConverter \"C:\\Users\\Kristian\\Downloads\\Anime\\[Kametsu] ERASED (BD 1080p Hi10 FLAC)\""));
        //System.exit(1);

        int choice = getChoice("Which converter do you want do use?\n1. Anime to web mp4\n2. Audio converter\n3. VideoStream converter", 1, 3);

        printl("Input for this converter:");
        switch (choice) {
            case 1:
                animeConverter();
                break;
            case 2:
                con = new AudioConverter(FFPROBE_PATH, FFMPEG_PATH, getChoice("<output extension>"));
                break;
            case 3:
                con = new VideoConverter(FFPROBE_PATH, FFMPEG_PATH, getChoice("<output extension>"));
                break;
            default:
                System.exit(1);
        }

        int recursionSteps = 1;

        printl("<Folder/File> [Recursions]: ");
        List<String> input = readInput(2);
        while (input.isEmpty()) {
            print("File path required.");
            input = readInput(2);
        }
        File folder = new File(input.get(0));
        if (input.size() > 1) {
            try {
                recursionSteps = Integer.parseInt(input.get(1));
            } catch (NumberFormatException e) {
                printl("Recursion steps is invalid and will be ignored.");
            }
        }

        if (folder.isDirectory()) {
            File[] files = listFilesRec(folder, con.getValidFormats(), recursionSteps);
            if (files != null && files.length > 0) {
                for (File file : files) {
                    con.convert(file);
                }
            } else {
                printl("No valid files found in folder.");
            }
        } else if (folder.exists()) {
            con.convert(folder);
        } else {
            System.out.println("Path " + folder.getAbsolutePath() + " does not point to any file or folder.");
        }
        WRITER.close();
    }
    
    /**
     * Prints a string
     * @param input <p>The string to print.</p>
     * @throws IOException <p>If the writer fails to write.</p>
     */
    private static void print(String input) throws IOException {
       WRITER.write(input);
       WRITER.flush();
    }
    
    /**
     * Prints a string and a line break
     * @param input <p>The string to print.</p>
     * @throws IOException <p>If the writer fails to write.</p>
     */
    private static void printl(String input) throws IOException {
       WRITER.write(input);
       WRITER.newLine();
       WRITER.flush();
    }

    private enum converterArgumentValueType {
        BOOLEAN,
        COMMA_SEPARATED_LIST,
        SINGLE_VALUE,
        INT
    }

    private static class converterArgument {
        private String name;
        private boolean valueRequired;
        private converterArgumentValueType valueType;
        private converterArgument(String name, boolean valueRequired, converterArgumentValueType valueType) {
            this.name = name;
            this.valueRequired = valueRequired;
            this.valueType = valueType;
        }

        private boolean testArgumentValue(String value) {
            if (value.length() == 0) {
                return !valueRequired;
            }
            if (valueRequired && value.startsWith("-")) {
                return false;
            }
            switch (valueType) {
                case BOOLEAN:
                    String lower = value.toLowerCase();
                    return lower.equals("true") || lower.equals("false");
                case COMMA_SEPARATED_LIST:
                    return !value.contains(" ");
                case SINGLE_VALUE:
                    return !value.contains(" ");
                case INT:
                    int ignored = Integer.parseInt(value);
                    return true;
            }
            return false;
        }
    }

    private static void parser(List<String> tokens) {
        String[] types = {"animeconverter", "audioconverter", "videoconverter"};
        converterArgument[] commonArgs = {
                new converterArgument("-recursions", true, converterArgumentValueType.INT)
        };
        converterArgument[] animeArgs = {

        };
        converterArgument[] audioArgs = {
                new converterArgument("-outext", true, converterArgumentValueType.SINGLE_VALUE)
        };
        converterArgument[] videoArgs = {
                new converterArgument("-outext", true, converterArgumentValueType.SINGLE_VALUE)
        };
        String type = tokens.get(0).toLowerCase();
        if (!listContains(types, s -> s.equals(type))) {
            throw new IllegalArgumentException("Unknown converter type chosen.");
        }
        if (tokens.size() < 2) {
            throw new IllegalArgumentException("No file/folder path in argument.");
        }
        for (int i = 1; i < tokens.size() - 1; i++) {
            //TODO: Find the type of argument and check the value
            //TODO: Find an executable way to represent the chain of commands parsed
        }
    }

    /**
     * Tokenizes a string
     * @param input <p>A string.</p>
     * @return <p>A list of tokens.</p>
     */
    private static List<String> tokenizer(String input) {
        List<String> tokens = new ArrayList<>();
        boolean startedQuote = false;
        StringBuilder currentToken = new StringBuilder();
        for (int i = 0; i < input.length(); i++) {
            char c = input.charAt(i);
            switch (c) {
                case ' ':
                    if (!startedQuote) {
                        //If not inside "", a space marks the end of a parameter
                        if (!currentToken.toString().trim().equals("")) {
                            tokens.add(currentToken.toString());
                            currentToken = new StringBuilder();
                        } else {
                            currentToken = new StringBuilder();
                        }
                    } else {
                        currentToken.append(c);
                    }
                    break;
                case '"':
                    if (startedQuote) {
                        if (!currentToken.toString().trim().equals("")) {
                            tokens.add(currentToken.toString());
                            currentToken = new StringBuilder();
                        }
                        startedQuote = false;
                    } else {
                        startedQuote = true;
                        currentToken = new StringBuilder();
                    }   break;
                default:
                    currentToken.append(c);
                    if (i == input.length() - 1) {
                        tokens.add(currentToken.toString());
                    }   break;
            }
        }
        return tokens;
    }

    /**
     * Initializes the anime converter
     * @throws IOException <p>If reading or writing fails.</p>
     */
    private static void animeConverter() throws IOException {
        printl("[Audio languages jpn,eng,ger,fre] [Subtitle languages eng,ger,fre] [Convert to stereo if necessary true/false] [Prevent signs&songs subtitles true/false]\nYour input: ");
        List<String> input = readInput(4);
        String[] audioLang = new String[]{"jpn", "*"};
        String[] subtitleLang = new String[]{"eng", "*"};
        boolean toStereo = true;
        boolean preventSigns = true;
        if (input.size() > 0 && getList(input, 0) != null) {
            audioLang = getList(input, 0);
        }
        if (input.size() > 1 && getList(input, 1) != null) {
            subtitleLang = getList(input, 1);
        }
        if (input.size() > 2) {
            toStereo = Boolean.parseBoolean(input.get(2));
        }
        if (input.size() > 3) {
            preventSigns = Boolean.parseBoolean(input.get(3));
        }
        con = new AnimeConverter(FFPROBE_PATH, FFMPEG_PATH, audioLang, subtitleLang, toStereo, preventSigns);
    }

    /**
     * Gets a list from a comma separated string at index in list
     * @param list <p>A list of tokens.</p>
     * @param index <p>The index of the token containing comma separated entries.</p>
     * @return <p>A string list.</p>
     */
    private static String[] getList(List<String> list, int index) {
        String[] result = null;
        if (list.size() > index) {
            if (list.get(index).contains(",")) {
                result = list.get(index).split(",");
            } else {
                result = new String[]{list.get(index)};
            }
        }
        return result;
    }

    /**
     * Reads a number of tokens from the user input
     * @param max <p>The number of tokens expected.</p>
     * @return <p>A list of tokens.</p>
     */
    private static List<String> readInput(int max) {
        List<String> tokens =  tokenizer(READER.nextLine());
        if (max < tokens.size()) {
            throw new IllegalArgumentException("Input contains " + tokens.size() +
                    " arguments, but the input only supports " + max + " arguments.");
        }
        return tokens;
    }

    /**
     * Gets the user's choice
     * @param prompt <p>The prompt shown to the user.</p>
     * @return <p>The non-empty choice given by the user.</p>
     * @throws IOException <p>If reading or writing fails.</p>
     */
    private static String getChoice(String prompt) throws IOException {
        printl(prompt);
        String choice = "";
        while (choice.equals("")) {
            printl("Your input: ");
            choice = READER.nextLine();
        }
        return choice;
    }

    /**
     * Gets an integer from the user
     * @param prompt    The prompt to give the user
     * @param min       The minimum allowed value
     * @param max       The maximum allowed value
     * @return          The value given by the user
     */
    private static int getChoice(String prompt, int min, int max) throws IOException {
        printl(prompt);
        int choice = 0;
        while (choice < min || choice > max) {
            printl("Your input: ");
            try {
                choice = Integer.parseInt(READER.next());
            } catch (NumberFormatException e) {
                printl("Invalid choice. Please try again.");
            } finally {
                READER.nextLine();
            }
        }
        return choice;
    }

    /**
     * Tests if any element in a list fulfills a condition.
     *
     * @param list      The list to test against
     * @param predicate A predicate to use on every element in the list
     * @param <T>       Anything which can be stored in a list
     * @return          True if at least one element fulfills the predicate
     */
    private static <T> boolean listContains(T[] list, Predicate<T> predicate) {
        for (T item : list) {
            if (predicate.test(item)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Recursively lists all files in a folder
     *
     * @param folder    The folder to start from
     * @param maxRec    Maximum number of recursions
     * @return          A list of files
     */
    private static File[] listFilesRec(File folder, String[] extensions, int maxRec) {
        if (maxRec == 0) { return null; }
        File[] listOfFiles = folder.listFiles((file) -> file.isFile() && listContains(extensions, (item) -> file.getName().endsWith(item)));
        if (listOfFiles == null) { return null; }
        if (maxRec > 1) {
            File[] listOfFolders = folder.listFiles((dir, name) -> new File(dir, name).isDirectory());
            if (listOfFolders != null) {
                for (File file : listOfFolders) {
                    File[] nextLevel = listFilesRec(file, extensions, maxRec - 1);
                    if (nextLevel != null) {
                        listOfFiles = Converter.concatenate(listOfFiles, nextLevel);
                    }
                }
            }
        }
        return listOfFiles;
    }
}
